import json
import os
import time
from functools import lru_cache
from typing import Dict, List

import hubble
from docarray import Document, DocumentArray
from hubble.excepts import AuthenticationRequiredError
from jina import Executor, requests
from jina.enums import LogVerbosity
from jina.logging.logger import JinaLogger

from now.now_dataclasses import UserInput


class SecurityLevel:
    ADMIN = 1
    USER = 2


def secure_request(level: int, on: str = None):
    """decorator to check the authorization of the incoming request"""

    def decorator(func):
        @requests(on=on)
        def wrapper(*args, **kwargs):
            _check_user(
                kwargs,
                level,
                args[0].user_emails,
                args[0].admin_emails,
                args[0].api_keys,
            )

            cls_instance = args[0]
            if 'args' in kwargs:
                cls_instance.logger.debug(f"args: {kwargs['args']}")

            if 'docs' in kwargs and kwargs['docs']:
                if isinstance(kwargs['docs'], DocumentArray):
                    cls_instance.logger.info(f"processing {len(kwargs['docs'])} docs")
                    if cls_instance.logger.logger.level <= LogVerbosity.DEBUG:
                        kwargs['docs'].summary()
                        kwargs['docs'][0].summary()
                elif isinstance(kwargs['docs'], Document):
                    cls_instance.logger.info(f"processing a single doc")
                    if cls_instance.logger.logger.level <= LogVerbosity.DEBUG:
                        kwargs['docs'].summary()

            if 'docs_matrix' in kwargs and kwargs['docs_matrix']:
                cls_instance.logger.info(
                    f"processing {len(kwargs['docs_matrix'])} docs_matrix"
                )
                lengths = [str(len(da)) for da in kwargs['docs_matrix']]
                cls_instance.logger.info(f"lengths: {', '.join(lengths)}")
                if cls_instance.logger.logger.level <= LogVerbosity.DEBUG and len(
                    kwargs['docs_matrix']
                ):
                    kwargs['docs_matrix'][0].summary()
                    kwargs['docs_matrix'][0][0].summary()

            if 'docs_map' in kwargs and kwargs['docs_map']:
                cls_instance.logger.info(
                    f"processing {len(kwargs['docs_map'])} docs_map"
                )
                lengths = [
                    f"{key_encoder}: {len(da)}"
                    for key_encoder, da in kwargs['docs_map'].items()
                ]
                cls_instance.logger.info(f"lengths: {', '.join(lengths)}")

            t0 = time.time()
            res = func(*args, **kwargs)
            t1 = time.time()

            if isinstance(res, DocumentArray):
                cls_instance.logger.info(
                    f"output {len(res)} docs in {(t1 - t0) * 1000:.1f}ms"
                )
                if len(res) and cls_instance.logger.logger.level <= LogVerbosity.DEBUG:
                    res[0].summary()
            elif not res:
                cls_instance.logger.info(f"no output in {(t1 - t0) * 1000:.1f}ms")

            return res

        return wrapper

    return decorator


def _check_user(kwargs, level, user_emails, admin_emails, api_keys):
    if user_emails == [] and admin_emails == [] and api_keys == []:
        return

    if 'parameters' not in kwargs or (
        'api_key' not in kwargs['parameters'] and 'jwt' not in kwargs['parameters']
    ):
        raise PermissionError(
            f'`jwt` or `api_key` needs to be part of the request parameters.'
        )

    if 'api_key' in kwargs['parameters']:
        if kwargs['parameters']['api_key'] in api_keys:
            return
        else:
            raise PermissionError(f'`api_key` is invalid')
    else:
        jwt = kwargs['parameters']['jwt']

    user_info = _get_user_info(jwt['token'])
    for email in admin_emails + user_emails + ['jina.ai']:
        if _valid_user(user_info.get('email'), email):
            if level == SecurityLevel.ADMIN and email not in admin_emails:
                raise PermissionError(f'User {email} is not an admin.')
            return
    raise PermissionError(
        f'User {user_info.get("email") or user_info["_id"]} has no permission'
    )


def _valid_user(user_email, allowed_email):
    if '@' not in allowed_email:
        return user_email.split('@')[1] == allowed_email
    else:
        return user_email == allowed_email


@lru_cache(maxsize=128, typed=False)
def _get_user_info(token):
    try:
        client = hubble.Client(token=token, max_retries=None, jsonify=True)
        user_info = client.get_user_info()
        if user_info['code'] != 200:
            raise PermissionError(
                'Wrong token passed or the token is expired! Please re-login'
            )
        return user_info['data']
    except AuthenticationRequiredError as e:
        raise PermissionError(
            'Token expired or invalid. Please use the updated token', e
        )


def get_auth_executor_class():
    class NOWAuthExecutor(Executor):
        """
        NOWAuthExecutor performs the token check for authorization. It stores the owner ID belonging
        to the authorised user and also the `user_id` of the allowed users with access to the flow
        deployed by the user.

        If no `admin_emals`, `user_emails` and `api_keys` are provided, no checks will be performed.
        """

        def __init__(
            self,
            user_input_dict: Dict = {},
            admin_emails: List[str] = [],
            user_emails: List[str] = [],
            api_keys: List[str] = [],
            *args,
            **kwargs,
        ):
            """
            :param user_input_dict: kwargs to construct UserInput
            :param admin_email: ID of the user deploying this flow. ID is obtained from Hubble
            :param user_emails: Comma separated Email IDs of the allowed users with access to this flow.
                The Email ID from the incoming request to this flow will be verified against this.
            :param pats: List of PATs of the allowed users with access to this flow.
            """
            super().__init__(*args, **kwargs)
            self.logger = JinaLogger(self.__class__.__name__)
            self.admin_emails = admin_emails
            self.user_emails = user_emails
            self.api_keys = api_keys
            self._user = None

            self.user_input = UserInput()
            for attr_name, prev_value in self.user_input.__dict__.items():
                setattr(
                    self.user_input,
                    attr_name,
                    user_input_dict.get(attr_name, prev_value),
                )

            # TODO admin email must be persisted as well
            # TODO also, please remove duplicate code

            self.api_keys_path = (
                os.path.join(self.workspace, 'api_keys.json')
                if self.workspace
                else None
            )
            self.user_emails_path = (
                os.path.join(self.workspace, 'user_emails.json')
                if self.workspace
                else None
            )

            if self.api_keys_path and os.path.exists(self.api_keys_path):
                with open(self.api_keys_path, 'r') as fp:
                    self.api_keys = json.load(fp)
            if self.user_emails_path and os.path.exists(self.user_emails_path):
                with open(self.user_emails_path, 'r') as fp:
                    self.user_emails = json.load(fp)

        @secure_request(on='/admin/updateUserEmails', level=SecurityLevel.ADMIN)
        def update_user_emails(self, parameters: Dict = None, **kwargs):
            """
            Update the email addresses during runtime. That way, we don't have to restart it.
            """
            self.user_emails = parameters['user_emails']
            if self.user_emails_path:
                with open(self.user_emails_path, 'w') as fp:
                    json.dump(self.user_emails, fp)

        @secure_request(on='/admin/updateApiKeys', level=SecurityLevel.ADMIN)
        def update_api_keys(self, parameters: Dict = None, **kwargs):
            """
            Update the api keys during runtime. That way, we don't have to restart it.
            """
            self.api_keys = parameters['api_keys']
            if self.api_keys_path:
                with open(self.api_keys_path, 'w') as fp:
                    json.dump(self.api_keys, fp)

        @secure_request(level=SecurityLevel.USER)
        def check(self, docs: DocumentArray = None, **kwargs):
            """
            Check the authorization for each incoming request. The logic of the function is in the decorator.
            """
            return docs

    return NOWAuthExecutor



# TODO bff_request_mapping_fn and bff_response_mapping_fn should be used to create all routes

import sys

import uvicorn
from fastapi import FastAPI, Request
from fastapi.middleware.cors import CORSMiddleware
from starlette.applications import Starlette
from starlette.responses import JSONResponse

from now import __version__
from now.executor.gateway.bff.app.decorators import api_method, timed
from now.executor.gateway.bff.app.v1.routers.app import extras_router, search_app_router
from now.executor.gateway.bff.logger import logger

TITLE = 'Jina NOW'
DESCRIPTION = 'The Jina NOW service API'
AUTHOR = 'Jina AI'
EMAIL = 'hello@jina.ai'


def get_app_instance():
    """Build FastAPI app."""
    app = FastAPI(
        title=TITLE,
        description=DESCRIPTION,
        contact={
            'author': AUTHOR,
            'email': EMAIL,
        },
    )

    app.add_middleware(
        CORSMiddleware,
        allow_origins=['*'],
        allow_credentials=True,
        allow_methods=['*'],
        allow_headers=['*'],
    )

    @app.get('/ping')
    @api_method
    @timed
    def check_liveness() -> str:
        """
        Sanity check - this will let the caller know that the service is operational.
        """
        return 'pong!'

    @app.get('/')
    @api_method
    @timed
    def read_root() -> str:
        """
        Root path welcome message.
        """
        return (
            f'{TITLE} v{__version__} 🚀 {DESCRIPTION} ✨ '
            f'author: {AUTHOR} email: {EMAIL} 📄  '
            'Check out /docs or /redoc for the full API documentation!'
        )

    @app.exception_handler(Exception)
    async def unicorn_exception_handler(request: Request, exc: Exception):
        import traceback

        error = traceback.format_exc()
        return JSONResponse(
            status_code=500,
            content={
                "message": f"Exception in BFF, but the root cause can still be in the flow: {error}"
            },
        )

    return app


def build_app():
    # search app router
    search_app_mount = '/api/v1/search-app'
    search_app_app = get_app_instance()
    search_app_app.include_router(search_app_router)

    # Admin router
    admin_mount = '/api/v1/admin'
    admin_app = get_app_instance()
    admin_app.include_router(extras_router)

    sub_apps = {admin_mount: admin_app, search_app_mount: search_app_app}

    # Mount them - for other modalities just add an app instance
    app = Starlette()
    for route, sub_app in sub_apps.items():
        app.mount(route, sub_app)

    # startup event can only be fired for the parent app due to the lifecycle of the apps.
    # In order to add some code for mounted (or sub) apps to be executed on the startup
    # event, below is the hack

    @app.on_event('startup')
    def startup():
        for sub_app in sub_apps.values():
            logger.info(f'Jina NOW sub-app {sub_app} started! ')
            # add here app-specific startup code

    return app


application = build_app()


def run_server(port=8080):
    """Run server."""
    app = build_app()
    uvicorn.run(
        app,
        host='0.0.0.0',
        port=port,
        loop='uvloop',
        http='httptools',
    )


if __name__ == '__main__':
    try:
        run_server(8080)
    except Exception as exc:
        logger.critical(str(exc))
        logger.exception(exc)
        sys.exit(1)



from hubble import Client as HubbleClient  # noqa
from hubble.payment.client import PaymentClient  # noqa
from jina.logging.logger import JinaLogger

from now.constants import (
    NOWGATEWAY_SEARCH_FEE_PRO_QUANTITY,
    NOWGATEWAY_SEARCH_FEE_QUANTITY,
)
from now.executor.abstract.auth.auth import _get_user_info  # noqa
from now.executor.gateway.billing.client import BillingClient

logger = JinaLogger(__file__)


class UserPaymentClient:
    user_id: str = None
    user_payment_client = None  # type: BillingClient

    @classmethod
    def init_user_id(cls, user_id: str):
        logger.info('** Entered init_user_id() **')
        cls.user_id = user_id
        cls.user_payment_client = BillingClient(user_id=user_id)
        logger.info('** Exited init_user_id() **')


def report_search_usage(token=None):
    logger.info('** Entered report_search_usage() **')
    payment_client = UserPaymentClient.user_payment_client
    payment_client.report(
        quantity_basic=NOWGATEWAY_SEARCH_FEE_QUANTITY,
        quantity_pro=NOWGATEWAY_SEARCH_FEE_PRO_QUANTITY,
    )



import os
import traceback
from json import JSONDecodeError

from hubble.excepts import BaseError
from hubble.payment.client import PaymentClient  # noqa
from jina.logging.logger import JinaLogger

from now.utils.common.helpers import current_time

logger = JinaLogger(__file__)


class BillingClient:
    def __init__(self, user_id: str):
        self.user_id = user_id
        self.m2m_token = None
        self.payment_client = None
        self.session_token = None
        self.init_payment_client()

    def init_payment_client(self):
        try:
            logger.info('** Entered init_payment_client() **')
            self.m2m_token = os.environ.get('M2M')
            self.payment_client = PaymentClient(m2m_token=self.m2m_token)
            self.session_token = self.payment_client.get_user_token(
                user_id=self.user_id
            )['data']
        except (BaseError, JSONDecodeError) as e:
            logger.error(f'Error while initialising payment client for user, {e}')
            traceback.print_exc()
            raise e
        except Exception as e:  # noqa
            logger.error(
                f'Found an unhandled exception while initialising payment client, {e}'
            )
            traceback.print_exc()
            raise e

    def report(self, quantity_basic, quantity_pro):
        logger.info('Time of reporting for credits usage: {}'.format(current_time()))
        app_id = 'search'
        try:
            acc_status = self.get_account_status()
            product_id = acc_status['internal_product_id']
            logger.info(
                f'Credits before: {acc_status["credits"]} & product_id: {product_id}'
            )
            if product_id == 'free-plan':
                quantity = quantity_basic
            else:
                quantity = quantity_pro
            if self.can_charge(acc_status):
                self.payment_client.report_usage(  # noqa
                    self.session_token,
                    app_id,
                    product_id,
                    quantity,
                    meta={
                        'appId': 'search-app',
                        'feeType': 'search-fee',
                        'source': 'gateway',
                        'planType': product_id,
                        'flowId': os.environ.get('FLOW_ID'),
                        'flowNamespace': 'jnamespace-'
                        + os.environ.get('K8S_NAMESPACE_NAME', ''),
                        'timeNow': current_time(),
                    },
                )
                logger.info(f'**** `{round(quantity, 3)}` credits charged ****')
                acc_status = self.get_account_status()
                logger.info(f'Credits after: {acc_status["credits"]}')
            else:
                logger.info(f'**** Could not charge. Check payment summary ****')
            logger.info(f'User account status:\n{acc_status}')
        except Exception as e:
            # Do not continue with request if payment fails
            logger.error(f'Error while reporting credits usage {e}')
            traceback.print_exc()
            raise e

    def get_account_status(self):
        resp = self.payment_client.get_summary(
            token=self.session_token, app_id='search'
        )
        has_payment_method = resp['data'].get('hasPaymentMethod', False)
        user_credits = resp['data'].get('credits', None)
        internal_product_id = resp['data'].get('internalProductId', None)
        user_account_status = resp['data'].get('userAccountStatus', None)
        return {
            'has_payment_method': has_payment_method,
            'credits': user_credits,
            'internal_product_id': internal_product_id,
            'user_account_status': user_account_status,
        }

    @staticmethod
    def can_charge(user_acc_status):
        return (
            (user_acc_status['credits'] is not None and user_acc_status['credits'] > 0)
            or user_acc_status['has_payment_method']
        ) and user_acc_status['user_account_status'] == 'active'



import json
import logging
import os

from now.now_dataclasses import UserInput

logger = logging.getLogger(__name__)

ENV_PREFIX = 'JINA_NOW_'

# server
DEFAULT_WORKERS = 1
DEFAULT_PORT = 8080
DEFAULT_BACKLOG = 2048

# debug flag
DEFAULT_DEBUG = True


class GlobalUserInput:
    user_input_in_bff = UserInput()

    @classmethod
    def update(cls, user_input_dict: dict):
        for attr_name, prev_value in cls.user_input_in_bff.__dict__.items():
            setattr(
                cls.user_input_in_bff,
                attr_name,
                user_input_dict.get(attr_name, prev_value),
            )

    @classmethod
    def reset(cls):
        cls.user_input_in_bff = UserInput()


def init_user_input_in_bff():
    try:
        with open(os.path.join(os.path.expanduser('~'), 'user_input.json'), 'r') as f:
            user_input_dict = json.load(f)
            GlobalUserInput.update(user_input_dict)
    except FileNotFoundError:
        print('Could not find user input file in BFF')
        print(f'used path: {os.path.join(os.path.expanduser("~"), "user_input.json")}')
        print('but this can be okay')
        GlobalUserInput.reset()  # reset to default


init_user_input_in_bff()



import os.path
import tempfile
import time
from typing import Union

from jina.clients import Client
from jina.jaml import JAML

from now.deployment import deployment
from now.log import time_profiler
from now.utils.jcloud.helpers import read_flow_file, write_flow_file


@time_profiler
def deploy_flow(flow_yaml: Union[str, dict]):
    """
    Deploy a Flow on JCloud.
    :param flow_yaml: Path to a yaml file or a dict or a Yaml string.

    """
    # TODO create tmpdir top level and pass it down
    with tempfile.TemporaryDirectory() as tmpdir:
        if isinstance(flow_yaml, str) and os.path.exists(flow_yaml):
            flow_file = flow_yaml
            flow_name = read_flow_file(flow_file)['jcloud']['name']
        else:
            if isinstance(flow_yaml, str):
                flow_yaml = JAML.load(flow_yaml)
            flow_name = flow_yaml['jcloud']['name']
            flow_file = os.path.join(tmpdir, 'flow.yml')
            write_flow_file(flow_yaml, flow_file)

        # Now deploy the flow to JCloud
        try:
            flow = deployment.deploy_wolf(path=flow_file)
        except BaseException as e:  # noqa
            print(f'Failed to deploy flow. Deleting flow: `{flow_name}`')
            flows = deployment.list_all_wolf(status='Failed', namespace=flow_name)
            if not flows or len(flows) > 1:
                print(f'Found {len(flows)} `FAILED` flows with name {flow_name}.')
            else:
                flow = deployment.terminate_wolf(flow_id=flows[0]['id'])
                print(f'Successfully deleted flow: `{flows[0]["id"]}`')
            raise e

    time.sleep(10)  # sleep for some time before starting with indexing
    gateway_host_http = flow.endpoints['gateway (http)']
    client = Client(host=gateway_host_http)

    return client, gateway_host_http



import base64
import json
import os
import tempfile
from collections import Counter, defaultdict
from concurrent.futures import ThreadPoolExecutor
from typing import Dict, List, Tuple, Type, Union

from docarray import Document, DocumentArray
from docarray.dataclasses import is_multimodal
from docarray.document.mixins.helper import _uri_to_blob

from now.common.detect_schema import (
    get_first_file_in_folder_structure_s3,
    get_s3_bucket_and_folder_prefix,
)
from now.constants import MAX_DOCS_FOR_BENCHMARKING, MAX_DOCS_FOR_TESTING, DatasetTypes
from now.data_loading.create_dataclass import create_dataclass
from now.data_loading.elasticsearch import ElasticsearchExtractor
from now.executor.preprocessor.s3_download import download_from_bucket
from now.log import yaspin_extended
from now.now_dataclasses import UserInput
from now.utils.common.helpers import flatten_dict, null_context, sigmap
from now.utils.docarray.helpers import get_chunk_by_field_name

MAX_WORKERS = 10


def load_data(user_input: UserInput, print_callback=print) -> DocumentArray:
    """Based on the user input, this function will pull the configured DocumentArray dataset ready for the preprocessing
    executor.

    :param user_input: The configured user object. Result from the Jina Now cli dialog.
    :param print_callback: The callback function that should be used to print the status.
    :return: The loaded DocumentArray.
    """
    da = None
    if user_input.dataset_type in [DatasetTypes.DEMO, DatasetTypes.DOCARRAY]:
        user_input.field_names_to_dataclass_fields = {
            field: field for field in user_input.index_fields
        }
        data_class = None
    else:
        data_class, user_input.field_names_to_dataclass_fields = create_dataclass(
            user_input=user_input
        )
    if user_input.dataset_type in [DatasetTypes.DOCARRAY, DatasetTypes.DEMO]:
        print_callback('⬇  Pull DocumentArray dataset')
        da = _pull_docarray(user_input.dataset_name, user_input.admin_name)
        da = _update_fields_and_metadata(da, user_input)
    elif user_input.dataset_type == DatasetTypes.PATH:
        print_callback('💿  Loading files from disk')
        da = _load_from_disk(user_input=user_input, data_class=data_class)
    elif user_input.dataset_type == DatasetTypes.S3_BUCKET:
        print_callback('🗄  Loading files from S3')
        da = list_files_from_s3_bucket(user_input=user_input, data_class=data_class)
    elif user_input.dataset_type == DatasetTypes.ELASTICSEARCH:
        print_callback('🔍  Loading data from Elasticsearch')
        da = _extract_es_data(user_input=user_input, data_class=data_class)
    da = set_modality_da(da)
    _convert_da_tensors_to_blob(da)
    _add_metadata_to_chunks(da, user_input)
    if da is None:
        raise ValueError(
            f'Could not load DocumentArray dataset. Please check your configuration: {user_input}.'
        )
    if 'NOW_CI_RUN' in os.environ:
        da = da[:MAX_DOCS_FOR_TESTING]
    elif 'NOW_BENCHMARK_RUN' in os.environ:
        da = da[:MAX_DOCS_FOR_BENCHMARKING]
    return da


def _add_metadata_to_chunks(da, user_input):
    dataclass_fields_to_field_names = {
        v: k for k, v in user_input.field_names_to_dataclass_fields.items()
    }
    for doc in da:
        for dataclass_field, meta_dict in doc._metadata['multi_modal_schema'].items():
            field_name = dataclass_fields_to_field_names.get(dataclass_field, None)
            if 'position' in meta_dict:
                get_chunk_by_field_name(doc, dataclass_field)._metadata[
                    'field_name'
                ] = field_name


def _is_valid_uri(uri):
    if uri.startswith('s3://'):
        return True
    if os.path.isfile(uri):  # for local files, we keep the blob info
        return False
    try:
        _uri_to_blob(uri)
        return True
    except:  # noqa
        return False


def _convert_document_tensors_to_blob(doc):
    for chunk in doc.chunks:
        # Document can have uri and (tensor or blob or text). Precedence: uri > blob > tensor
        if chunk.uri and (chunk.tensor is not None or chunk.blob):
            # If uri is not valid then unset it. Convert tensor to blob, if needed. Finally, unset tensor attribute.
            if not _is_valid_uri(chunk.uri):
                chunk.uri = ''  # if not a valid uri then unset it
                if not chunk.blob:
                    chunk.convert_image_tensor_to_blob()
                continue
            chunk.tensor = None  # unset tensor attribute
            chunk.blob = b''  # unset blob attribute


def _convert_da_tensors_to_blob(documents):
    with ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:
        for doc in documents:
            executor.submit(
                _convert_document_tensors_to_blob,
                doc,
            )


def _update_fields_and_metadata(
    da: DocumentArray, user_input: UserInput
) -> DocumentArray:
    """Add selected index fields to da, add the tags, remove non-index chunks,
    and update multi modal schema."""
    if not da:
        return da
    all_fields = da[0]._metadata['multi_modal_schema'].keys()
    for doc in da:
        filtered_chunks = []
        filtered_chunk_names = []
        for field in all_fields:
            field_doc = get_chunk_by_field_name(doc, field)
            if field not in user_input.index_fields:
                if field_doc.blob or field_doc.tensor is not None:
                    continue
                doc.tags.update(
                    {
                        field: field_doc.content
                        if isinstance(field_doc.content, str)
                        else field_doc.uri
                    }
                )
            else:
                filtered_chunks.append(field_doc)
                filtered_chunk_names.append(field)
        doc.chunks = filtered_chunks
        # keep only the index fields in metadata
        doc._metadata['multi_modal_schema'] = {
            field: doc._metadata['multi_modal_schema'][field]
            for field in filtered_chunk_names
        }
        # Update the positions accordingly to access the chunks
        for position, field in enumerate(filtered_chunk_names):
            doc._metadata['multi_modal_schema'][field]['position'] = int(position)

    return da


def _pull_docarray(dataset_name: str, admin_name: str) -> DocumentArray:
    dataset_name = (
        admin_name + '/' + dataset_name if '/' not in dataset_name else dataset_name
    )
    try:
        docs = DocumentArray.pull(
            name=dataset_name, show_progress=True, local_cache=True
        )
        if is_multimodal(docs[0]):
            return docs
        else:
            raise ValueError(
                f'The dataset {dataset_name} does not contain a multimodal DocumentArray. '
                f'Please check documentation https://docarray.jina.ai/fundamentals/dataclass/construct/'
            )
    except Exception:
        raise ValueError(
            'DocumentArray does not exist or you do not have access to it. '
            'Make sure to add user name as a prefix. Check documentation here. '
            'https://docarray.jina.ai/fundamentals/cloud-support/data-management/'
        )


def _extract_es_data(user_input: UserInput, data_class: Type) -> DocumentArray:
    query = {
        'query': {'match_all': {}},
        '_source': True,
    }
    es_extractor = ElasticsearchExtractor(
        query=query,
        index=user_input.es_index_name,
        user_input=user_input,
        data_class=data_class,
        connection_str=user_input.es_host_name,
    )
    extracted_docs = es_extractor.extract()
    return extracted_docs


def _load_from_disk(user_input: UserInput, data_class: Type) -> DocumentArray:
    """
    Loads the data from disk into multimodal documents.

    :param user_input: The user input object.
    :param data_class: The dataclass to use for the DocumentArray.
    """
    dataset_path = user_input.dataset_path.strip()
    dataset_path = os.path.expanduser(dataset_path)
    if os.path.isfile(dataset_path):
        try:
            da = DocumentArray.load_binary(dataset_path)
            if is_multimodal(da[0]):
                da = _update_fields_and_metadata(da, user_input)
                return da
            else:
                raise ValueError(
                    f'The file {dataset_path} does not contain a multimodal DocumentArray.'
                    f'Please check documentation https://docarray.jina.ai/fundamentals/dataclass/construct/'
                )
        except Exception:
            print(f'Failed to load the binary file provided under path {dataset_path}')
            exit(1)
    elif os.path.isdir(dataset_path):
        with yaspin_extended(
            sigmap=sigmap, text="Loading data from folder", color="green"
        ) as spinner:
            spinner.ok('🏭')
            docs = from_files_local(
                dataset_path,
                user_input.index_fields,
                user_input.field_names_to_dataclass_fields,
                data_class,
            )
            return docs
    else:
        raise ValueError(
            f'The provided dataset path {dataset_path} does not'
            f' appear to be a valid file or folder on your system.'
        )


def from_files_local(
    path: str,
    fields: List[str],
    field_names_to_dataclass_fields: Dict,
    data_class: Type,
) -> DocumentArray:
    """Creates a Multi Modal DocumentArray over a list of file path or the content of the files.

    :param path: The path to the directory
    :param fields: The fields to search for in the directory
    :param field_names_to_dataclass_fields: The mapping of the field names to the dataclass fields
    :param data_class: The dataclass to use for the document

    :return: A DocumentArray with the documents
    """

    file_paths = []
    for root, dirs, files in os.walk(path):
        file_paths.extend(
            [os.path.join(root, file) for file in files if not file.startswith('.')]
        )
    folder_generator = os.walk(path, topdown=True)
    current_level = folder_generator.__next__()
    folder_structure = 'sub_folders' if len(current_level[1]) > 0 else 'single_folder'
    if folder_structure == 'sub_folders':
        docs, _ = create_docs_from_subdirectories(
            file_paths, fields, field_names_to_dataclass_fields, data_class
        )
    else:
        docs, _ = create_docs_from_files(
            file_paths, fields, field_names_to_dataclass_fields, data_class
        )
    return DocumentArray(docs)


def create_docs_from_subdirectories(
    file_paths: List,
    fields: List[str],
    field_names_to_dataclass_fields: Dict,
    data_class: Type,
    path: str = None,
    is_s3_dataset: bool = False,
    download_content: bool = False,
    bucket=None,
) -> Tuple[List[Document], Dict[str, dict]]:
    """
    Creates a Multi Modal DocumentArray over a list of subdirectories.

    :param file_paths: The list of file paths
    :param fields: The fields to search for in the directory
    :param field_names_to_dataclass_fields: The mapping of the field names to the dataclass fields
    :param data_class: The dataclass to use for the document
    :param path: The path to the directory
    :param is_s3_dataset: Whether the dataset is stored on s3
    :param download_content: Whether to download the content of the s3 files
    :param bucket: The s3 bucket instance

    :return: The list of documents
    """

    docs = []
    non_json_file_info = []
    json_file_info = []
    folder_files = defaultdict(list)
    folder_files_content = defaultdict(dict)
    # Group the files by the folder they are in
    for file in file_paths:
        path_to_last_folder = (
            '/'.join(file.split('/')[:-1])
            if is_s3_dataset
            else os.sep.join(file.split(os.sep)[:-1])
        )
        folder_files[path_to_last_folder].append(file)

    # Process each files in each folder
    for folder, files in folder_files.items():
        kwargs = {}
        tags_loaded_local = {}
        _s3_uri_for_tags = ''
        file_info = [
            _extract_file_and_full_file_path(file, path, is_s3_dataset)
            for file in files
        ]

        if download_content:
            for file_name, file_full_path in file_info:
                content = get_file_content(bucket, file_full_path, is_s3_dataset)
                folder_files_content[folder.split('/')[-1]][file_name] = content

        for f_name, f_path in file_info:
            if f_name.endswith('.json'):
                json_file_info.append((f_name, f_path))
            else:
                non_json_file_info.append((f_name, f_path))

        # first store index fields given as non json files
        for file, file_full_path in non_json_file_info:
            if file in fields:
                kwargs[field_names_to_dataclass_fields[file]] = file_full_path

        # next check json files that can also contain index fields, and carry on data
        for file, file_full_path in json_file_info:
            if is_s3_dataset:  # s3 data source
                _s3_uri_for_tags = file_full_path
                for field in data_class.__annotations__.keys():
                    if field not in kwargs.keys():
                        kwargs[field] = file_full_path
            else:  # local data source
                with open(file_full_path) as f:
                    json_data = flatten_dict(json.load(f))
                for field, value in json_data.items():
                    if field in fields:
                        kwargs[field_names_to_dataclass_fields[field]] = value
                    else:
                        tags_loaded_local[field] = value
        doc = Document(data_class(**kwargs))
        if _s3_uri_for_tags:
            doc._metadata['_s3_uri_for_tags'] = _s3_uri_for_tags
        elif tags_loaded_local:
            doc.tags.update(tags_loaded_local)
        docs.append(doc)

    return docs, folder_files_content


def create_docs_from_files(
    file_paths: List,
    fields: List[str],
    field_names_to_dataclass_fields: Dict,
    data_class: Type,
    path: str = None,
    is_s3_dataset: bool = False,
    download_content: bool = False,
    bucket=None,
) -> Tuple[List[Document], Dict[str, dict]]:
    """
    Creates a Multi Modal DocumentArray over a list of files.

    :param file_paths: List of file paths
    :param fields: The fields to search for in the directory
    :param field_names_to_dataclass_fields: The mapping of the files to the dataclass fields
    :param data_class: The dataclass to use for the document
    :param path: The path to the directory
    :param is_s3_dataset: Whether the dataset is stored on s3
    :param download_content: Whether to download the content of the files and return as base64 encoded bytes
    :param bucket: The s3 bucket instance

    :return: A list of documents
    """
    docs = []
    folder_content = defaultdict()
    for file in file_paths:
        kwargs = {}
        file, file_full_path = _extract_file_and_full_file_path(
            file, path, is_s3_dataset
        )
        if download_content:
            folder_content[file] = get_file_content(
                bucket, file_full_path, is_s3_dataset
            )
        file_extension = file.split('.')[-1]
        if (
            file_extension == fields[0].split('.')[-1]
        ):  # fields should have only one index field in case of files only
            kwargs[field_names_to_dataclass_fields[fields[0]]] = file_full_path
            docs.append(Document(data_class(**kwargs)))
    return docs, folder_content


def _list_s3_file_paths(bucket, folder_prefix, top_n=None):
    """
    Lists the s3 file paths in an optimized way by finding the best level to use concurrent calls on
    in the file structure, using a threadpool.

    :param bucket: The s3 bucket used
    :param folder_prefix: The root folder prefix

    :return: A list of all s3 paths
    """
    # TODO bucket is not thread safe and outputs duplicate files for different prefixes
    # first_file = get_first_file_in_folder_structure_s3(bucket, folder_prefix)
    # structure_identifier = first_file[len(folder_prefix) :].split('/')
    # folder_structure = (
    #     'sub_folders' if len(structure_identifier) > 1 else 'single_folder'
    # )
    #
    # def get_level_order_prefixes(folder_prefix, level=1):
    #     """
    #     Gets the list of prefixes in a specific level. Levels are defined by the folder structure as follows:
    #     level_1/level_2/.../level_n/file.ext
    #
    #     :param folder_prefix: The current level prefix
    #     :param level: The desired level we want to get to
    #
    #     :return: A list of prefixes
    #     """
    #     level_prefixes = [
    #         obj['Prefix']
    #         for obj in bucket.meta.client.list_objects(
    #             Bucket=bucket.name, Prefix=folder_prefix, Delimiter='/'
    #         )['CommonPrefixes']
    #     ]
    #     if level == 1:
    #         return level_prefixes
    #     else:
    #         prefix_list = []
    #         for prefix in level_prefixes:
    #             prefix_list += get_level_order_prefixes(prefix, level - 1)
    #     return prefix_list
    #
    # def get_prefixes(max_levels=len(structure_identifier) - 2):
    #     """
    #     Finds the best level for the prefixes
    #
    #     :param max_levels: The maximum number of level we can get to, this defaults to len(structure_identifier) - 2
    #     because the latest level (len(structure_identifier) - 1) will only have files, so it won't have any common
    #     prefixes inside.
    #
    #     :return: A list of prefixes
    #     """
    #     level = 1
    #     list_prefixes = get_level_order_prefixes(folder_prefix, level)
    #     prefixes_states = [list_prefixes]
    #     while level < max_levels and len(list_prefixes) < NUM_FOLDERS_THRESHOLD:
    #         level += 1
    #         list_prefixes = get_level_order_prefixes(folder_prefix, level)
    #         prefixes_states.append(list_prefixes)
    #     if len(list_prefixes) > NUM_FOLDERS_THRESHOLD and len(prefixes_states) > 1:
    #         return prefixes_states[-2]
    #     return list_prefixes
    #
    # objects = []
    # if folder_structure == 'sub_folders':
    #     prefixes = get_prefixes()
    #     # TODO: change cpu count to a fixed number
    #     with ThreadPoolExecutor(max_workers=20) as executor:
    #         futures = []
    #         for prefix in prefixes:
    #             pref = ''.join(prefix)
    #             f = executor.submit(
    #                 lambda: list(bucket.objects.filter(Prefix=f'{pref}'))
    #             )
    #             futures.append(f)
    #         for f in futures:
    #             objects += f.result()
    # else:
    #     objects = list(bucket.objects.filter(Prefix=folder_prefix))
    if top_n:
        objects = list(bucket.objects.filter(Prefix=folder_prefix).limit(top_n))
    else:
        objects = list(bucket.objects.filter(Prefix=folder_prefix))
    return [
        obj.key
        for obj in objects
        if not obj.key.endswith('/') and not obj.key.split('/')[-1].startswith('.')
    ]


def validate_shrink_data(file_paths):
    """
    Validates the file paths: Keeps only consistent folders based on the number of files inside them.

    :param file_paths: The initial file paths.

    :return: The file paths to keep.
    """
    dict_fp = defaultdict(int)
    # take the folders of the files
    for file_path in file_paths:
        dict_fp['/'.join(file_path.split('/')[:-1])] += 1

    # remove the keys that don't have the same length as the majority
    val_counts = Counter(dict_fp.values())
    max_occ = max(list(val_counts.values()))
    max_common_element = 0
    for key, value in val_counts.items():
        if value == max_occ:
            max_common_element = max(max_common_element, key)
    keys_to_remove = []
    for key, val in dict_fp.items():
        if val != max_common_element:
            keys_to_remove.append(key)

    # keep only the desired file paths
    files_to_keep = []
    for file_path in file_paths:
        keep = True
        for key in keys_to_remove:
            if file_path.startswith(key):
                keep = False
                break
        if keep:
            files_to_keep.append(file_path)

    return files_to_keep


def get_file_content(bucket, uri, is_s3_uri=True):
    if is_s3_uri:
        with tempfile.TemporaryDirectory() as tmpdir:
            local_file = download_from_bucket(tmpdir, uri, bucket)

            with open(local_file, 'rb') as file:
                content = base64.b64encode(file.read())
    else:
        with open(uri, 'rb') as file:
            content = file.read()

    return content


def list_files_from_s3_bucket(
    user_input: UserInput,
    data_class: Type,
    top_n: int = None,
    verbose=True,
    return_file_content=False,
) -> Union[DocumentArray, Tuple[DocumentArray, Dict[str, dict]]]:
    """
    Loads the data from s3 into multimodal documents. It also supports additional params to return the contents
    of the files.

    :param user_input: The user input object.
    :param data_class: The dataclass to use for the DocumentArray.
    :param top_n: The number of files to pull, if set to None pull all files.
                    top_n is only set when calling from JAC
    :param verbose: If set to True, will print the progress bar.
    :param return_file_content: If set to True, will return all the file content in base64 encoded bytes.

    :return: The DocumentArray with the documents.
    """
    context_manager = yaspin_extended if verbose else null_context
    bucket, folder_prefix = get_s3_bucket_and_folder_prefix(user_input)
    first_file = get_first_file_in_folder_structure_s3(bucket, folder_prefix)
    structure_identifier = first_file[len(folder_prefix) :].split('/')
    folder_structure = (
        'sub_folders' if len(structure_identifier) > 1 else 'single_folder'
    )

    with context_manager(
        sigmap=sigmap, text="Listing files from S3 bucket ...", color="green"
    ) as spinner:
        file_paths = _list_s3_file_paths(bucket, folder_prefix, top_n)
        spinner.ok('🏭')

    if top_n:
        file_paths = validate_shrink_data(file_paths)

    with context_manager(
        sigmap=sigmap, text="Creating docarray from S3 bucket files ...", color="green"
    ) as spinner:
        if folder_structure == 'sub_folders':
            docs, content = create_docs_from_subdirectories(
                file_paths,
                user_input.index_fields,
                user_input.field_names_to_dataclass_fields,
                data_class,
                user_input.dataset_path,
                is_s3_dataset=True,
                bucket=bucket,
                download_content=return_file_content,
            )
        else:
            docs, content = create_docs_from_files(
                file_paths,
                user_input.index_fields,
                user_input.field_names_to_dataclass_fields,
                data_class,
                user_input.dataset_path,
                is_s3_dataset=True,
                bucket=bucket,
                download_content=return_file_content,
            )
        spinner.ok('👝')

    if return_file_content:
        return DocumentArray(docs), content
    return DocumentArray(docs)


def _extract_file_and_full_file_path(file_path, path=None, is_s3_dataset=False):
    """
    Extracts the file name and the full file path from s3 object.

    :param file_path: The file path
    :param path: The path to the directory
    :param is_s3_dataset: Whether the dataset is stored on s3

    :return: The file name and the full file path
    """
    if is_s3_dataset:
        file = file_path.split('/')[-1]
        file_full_path = '/'.join(path.split('/')[:3]) + '/' + file_path
    else:
        file_full_path = file_path
        file = file_path.split(os.sep)[-1]
    return file, file_full_path


def _get_modality(document: Document):
    """
    Detect document's modality based on its `modality` or `mime_type` attributes.

    :param document: The document to detect the modality for.
    """

    modalities = ['text', 'image', 'video']
    if document.modality:
        return document.modality
    mime_type_class = document.mime_type.split('/')[0]
    if document.mime_type == 'application/json':
        return 'text'
    if mime_type_class in modalities:
        return mime_type_class
    document.summary()
    raise ValueError(f'Unknown modality')


def set_modality_da(documents: DocumentArray):
    """
    Set document's modality based on its `modality` or `mime_type` attributes.

    :param documents: The DocumentArray to set the modality for.

    :return: The DocumentArray with the modality set.
    """
    for doc in documents:
        for chunk in doc.chunks:
            chunk.modality = chunk.modality or _get_modality(chunk)
    return documents



import json
from concurrent.futures import ProcessPoolExecutor
from time import time
from typing import Dict

import requests


def _call(http_host, request_body) -> float:
    t0 = time()
    resp = requests.post(f'{http_host}/api/v1/search-app/search', json=request_body)
    if resp.status_code != 200:
        raise Exception(
            f'Bad response code: {resp.status_code}\nContent: {resp.content}'
        )
    return time() - t0


def benchmark_deployment_latency(
    http_host: str,
    search_text: str,
    limit: int = 9,
    api_key: str = None,
    jwt: str = None,
    n_latency_calls: int = 30,
) -> Dict[str, float]:
    """Benchmarks a deployment by measuring latency on '/search' endpoint."""
    request_body = {
        'query': [{'name': 'query_text', 'modality': 'text', 'value': search_text}],
        'limit': limit,
    }
    if api_key:
        request_body['api_key'] = api_key
    if jwt:
        request_body['jwt'] = {'token': jwt}

    # measure latency
    start = time()
    for i in range(n_latency_calls):
        _call(http_host=http_host, request_body=request_body)
    return (time() - start) / n_latency_calls


def benchmark_deployment_qps_ps(
    http_host: str,
    search_text: str,
    limit: int = 9,
    api_key: str = None,
    jwt: str = None,
    n_qps_calls: int = 250,
    n_qps_workers: int = 10,
) -> Dict[str, float]:
    """Benchmarks a deployment by measuring QPS on '/search' endpoint."""
    request_body = {
        'query': [{'name': 'query_text', 'modality': 'text', 'value': search_text}],
        'limit': limit,
    }
    if api_key:
        request_body['api_key'] = api_key
    if jwt:
        request_body['jwt'] = {'token': jwt}

    result = {}
    # measure QPS
    with ProcessPoolExecutor(max_workers=n_qps_workers) as executor:
        # QPS test
        start = time()
        futures = []
        latencies = []

        for i in range(n_qps_calls):
            future = executor.submit(_call, http_host, request_body)
            futures.append(future)
        for future in futures:
            latencies.append(future.result())
        qps = n_qps_workers / (time() - start)
        result['qps'] = qps

    latencies = sorted(latencies)
    for p in [0, 50, 75, 85, 90, 95, 99, 99.9]:
        result[f'p{p}'] = latencies[int(len(latencies) * p / 100)]

    return result


if __name__ == '__main__':
    flow_name = ''
    api_key = ''
    search_text = ''
    limit = 60

    http_host = f'https://{flow_name}-http.wolf.jina.ai'
    latency = benchmark_deployment_latency(
        http_host=http_host, search_text=search_text, limit=limit, api_key=api_key
    )

    qps_dict = benchmark_deployment_qps_ps(
        http_host=http_host, search_text=search_text, limit=limit, api_key=api_key
    )

    print(json.dumps({'latency': latency, **qps_dict}, indent=4))
